import pytest
from hpcflow.api import (
    InputSourceType,
    Parameter,
    SchemaInput,
    SchemaOutput,
    TaskSchema,
    TaskSourceType,
    Task,
    InputValue,
    InputSource,
    InputSourceMode,
)


@pytest.fixture
def param_p1():
    return Parameter("p1")


@pytest.fixture
def param_p2():
    return Parameter("p2")


@pytest.fixture
def param_p3():
    return Parameter("p3")


def test_task_expected_input_source_mode_no_sources(param_p1):
    s1 = TaskSchema("ts1", actions=[], inputs=[SchemaInput(param_p1)])
    t1 = Task(
        schemas=s1,
        inputs=[InputValue(param_p1, value=101)],
    )
    assert t1.input_source_mode == InputSourceMode.AUTO


def test_task_expected_input_source_mode_with_sources(param_p1):
    s1 = TaskSchema("ts1", actions=[], inputs=[SchemaInput(param_p1)])
    t1 = Task(
        schemas=s1,
        inputs=[InputValue(param_p1, value=101)],
        input_sources=[InputSource.local()],
    )
    assert t1.input_source_mode == InputSourceMode.MANUAL


def test_task_get_available_task_input_sources_expected_return_first_task_local_value(
    param_p1,
):

    s1 = TaskSchema("ts1", actions=[], inputs=[SchemaInput(param_p1)])
    t1 = Task(schemas=s1, inputs=[InputValue(param_p1, value=101)])

    available = t1.get_available_task_input_sources()
    available_exp = {"p1": [InputSource(source_type=InputSourceType.LOCAL)]}

    assert available == available_exp


def test_task_get_available_task_input_sources_expected_return_first_task_default_value(
    param_p1,
):

    s1 = TaskSchema("ts1", actions=[], inputs=[SchemaInput(param_p1, default_value=101)])

    t1 = Task(schemas=s1)

    available = t1.get_available_task_input_sources()
    available_exp = {"p1": [InputSource(source_type=InputSourceType.DEFAULT)]}

    assert available == available_exp


def test_task_get_available_task_input_sources_expected_return_one_param_one_output(
    param_p1, param_p2
):

    s1 = TaskSchema(
        "ts1",
        actions=[],
        inputs=[SchemaInput(param_p1)],
        outputs=[SchemaOutput(param_p2)],
    )
    s2 = TaskSchema("ts2", actions=[], inputs=[SchemaInput(param_p2)])

    t1 = Task(schemas=s1, insert_ID=0)
    t2 = Task(schemas=s2)

    available = t2.get_available_task_input_sources([t1])
    available_exp = {
        "p2": [
            InputSource(
                source_type=InputSourceType.TASK,
                task_ref=0,
                task_source_type=TaskSourceType.OUTPUT,
            )
        ]
    }
    assert available == available_exp


def test_task_get_available_task_input_sources_expected_return_one_param_one_output_with_default(
    param_p1, param_p2
):

    s1 = TaskSchema(
        "ts1",
        actions=[],
        inputs=[SchemaInput(param_p1)],
        outputs=[SchemaOutput(param_p2)],
    )
    s2 = TaskSchema("ts2", actions=[], inputs=[SchemaInput(param_p2, default_value=2002)])

    t1 = Task(schemas=s1, insert_ID=0)
    t2 = Task(schemas=s2)

    available = t2.get_available_task_input_sources([t1])
    available_exp = {
        "p2": [
            InputSource(
                source_type=InputSourceType.TASK,
                task_ref=0,
                task_source_type=TaskSourceType.OUTPUT,
            ),
            InputSource(source_type=InputSourceType.DEFAULT),
        ]
    }
    assert available == available_exp


def test_task_get_available_task_input_sources_expected_return_one_param_one_output_with_local(
    param_p1, param_p2
):

    s1 = TaskSchema(
        "ts1",
        actions=[],
        inputs=[SchemaInput(param_p1)],
        outputs=[SchemaOutput(param_p2)],
    )
    s2 = TaskSchema("ts2", actions=[], inputs=[SchemaInput(param_p2)])

    t1 = Task(schemas=s1, insert_ID=0)
    t2 = Task(schemas=s2, inputs=[InputValue(param_p2, value=202)])

    available = t2.get_available_task_input_sources([t1])
    available_exp = {
        "p2": [
            InputSource(
                source_type=InputSourceType.TASK,
                task_ref=0,
                task_source_type=TaskSourceType.OUTPUT,
            ),
            InputSource(source_type=InputSourceType.LOCAL),
        ]
    }
    assert available == available_exp


def test_task_get_available_task_input_sources_expected_return_one_param_one_output_with_default_and_local(
    param_p1, param_p2
):

    s1 = TaskSchema(
        "ts1",
        actions=[],
        inputs=[SchemaInput(param_p1)],
        outputs=[SchemaOutput(param_p2)],
    )
    s2 = TaskSchema("ts2", actions=[], inputs=[SchemaInput(param_p2, default_value=2002)])

    t1 = Task(schemas=s1, insert_ID=0)
    t2 = Task(schemas=s2, inputs=[InputValue(param_p2, value=202)])

    available = t2.get_available_task_input_sources([t1])
    available_exp = {
        "p2": [
            InputSource(
                source_type=InputSourceType.TASK,
                task_ref=0,
                task_source_type=TaskSourceType.OUTPUT,
            ),
            InputSource(source_type=InputSourceType.LOCAL),
            InputSource(source_type=InputSourceType.DEFAULT),
        ]
    }
    assert available == available_exp


def test_task_get_available_task_input_sources_expected_return_one_param_two_outputs(
    param_p1, param_p2, param_p3
):
    s1 = TaskSchema(
        "ts1",
        actions=[],
        inputs=[SchemaInput(param_p1)],
        outputs=[SchemaOutput(param_p2), SchemaOutput(param_p3)],
    )
    s2 = TaskSchema(
        "ts2",
        actions=[],
        inputs=[SchemaInput(param_p2)],
        outputs=[SchemaOutput(param_p3), SchemaOutput("p4")],
    )
    s3 = TaskSchema("ts3", actions=[], inputs=[SchemaInput(param_p3)])

    t1 = Task(schemas=s1, insert_ID=0)
    t2 = Task(schemas=s2, insert_ID=1)
    t3 = Task(schemas=s3)

    available = t3.get_available_task_input_sources([t1, t2])
    available_exp = {
        "p3": [
            InputSource(
                source_type=InputSourceType.TASK,
                task_ref=0,
                task_source_type=TaskSourceType.OUTPUT,
            ),
            InputSource(
                source_type=InputSourceType.TASK,
                task_ref=1,
                task_source_type=TaskSourceType.OUTPUT,
            ),
        ]
    }
    assert available == available_exp


def test_task_get_available_task_input_sources_expected_return_two_params_one_output(
    param_p1,
    param_p2,
    param_p3,
):

    s1 = TaskSchema(
        "ts1",
        actions=[],
        inputs=[SchemaInput(param_p1)],
        outputs=[SchemaOutput(param_p2), SchemaOutput(param_p3)],
    )
    s2 = TaskSchema(
        "ts2",
        actions=[],
        inputs=[SchemaInput(param_p2), SchemaInput(param_p3)],
    )

    t1 = Task(schemas=s1, insert_ID=0)
    t2 = Task(schemas=s2)

    available = t2.get_available_task_input_sources([t1])
    available_exp = {
        "p2": [
            InputSource(
                source_type=InputSourceType.TASK,
                task_ref=0,
                task_source_type=TaskSourceType.OUTPUT,
            )
        ],
        "p3": [
            InputSource(
                source_type=InputSourceType.TASK,
                task_ref=0,
                task_source_type=TaskSourceType.OUTPUT,
            )
        ],
    }
    assert available == available_exp


def test_get_task_unique_names_two_tasks_no_repeats():
    s1 = TaskSchema("ts1", actions=[])
    s2 = TaskSchema("ts2", actions=[])

    t1 = Task(schemas=s1)
    t2 = Task(schemas=s2)

    assert Task.get_task_unique_names([t1, t2]) == ["ts1", "ts2"]


def test_get_task_unique_names_two_tasks_with_repeat():

    s1 = TaskSchema("ts1", actions=[])

    t1 = Task(schemas=s1)
    t2 = Task(schemas=s1)

    assert Task.get_task_unique_names([t1, t2]) == ["ts1_1", "ts1_2"]



# from multiprocessing import dummy
# import pytest
# from hpcflow.actions import Action
# from hpcflow.commands import Command

# from hpcflow.errors import InvalidIdentifier, MissingActionsError
# from hpcflow.task_schema import TaskObjective, TaskSchema


# @pytest.fixture
# def dummy_action():
#     return Action(commands=[Command("ls")])


# @pytest.fixture
# def dummy_schema(dummy_action):
#     return TaskSchema(TaskObjective("simulate"), actions=[dummy_action])


# @pytest.fixture
# def dummy_schema_args(dummy_action):
#     return {"objective": TaskObjective("simulate"), "actions": [dummy_action]}


# def test_init_with_str_objective(dummy_action):
#     obj_str = "simulate"
#     obj = TaskObjective(obj_str)
#     common = {"actions": [dummy_action]}
#     assert TaskSchema(obj_str, **common) == TaskSchema(obj, **common)


# def test_raise_on_missing_actions():
#     with pytest.raises(MissingActionsError):
#         TaskSchema("simulate", actions=[])


# def test_init_with_method_with_underscore(dummy_schema_args):
#     TaskSchema(method="my_method", **dummy_schema_args)


# def test_raise_on_invalid_method_digit(dummy_schema_args):
#     with pytest.raises(InvalidIdentifier):
#         TaskSchema(method="9", **dummy_schema_args)


# def test_raise_on_invalid_method_space(dummy_schema_args):
#     with pytest.raises(InvalidIdentifier):
#         TaskSchema(method="my method", **dummy_schema_args)


# def test_raise_on_invalid_method_non_alpha_numeric(dummy_schema_args):
#     with pytest.raises(InvalidIdentifier):
#         TaskSchema(method="_mymethod", **dummy_schema_args)


# def test_method_lowercasing(dummy_schema_args):
#     assert TaskSchema(method="MyMethod", **dummy_schema_args) == TaskSchema(
#         method="mymethod", **dummy_schema_args
#     )



# import pytest

# from hpcflow.parameters import Parameter, InputValue, ValueSequence
# from hpcflow.errors import InputValueDuplicateSequenceAddress


# def test_raise_on_duplicate_input_value_sequence_address():
#     p1 = Parameter("p1")
#     with pytest.raises(InputValueDuplicateSequenceAddress):
#         InputValue(
#             parameter=p1,
#             value={"A": 1},
#             sequences=[
#                 ValueSequence(values=[1, 2, 3], path=("A",), nesting_order=0),
#                 ValueSequence(values=[4, 5, 6], path=("A",), nesting_order=0),
#             ],
#         )


# def test_raise_on_duplicate_input_value_sequence_address_empty():
#     p1 = Parameter("p1")
#     with pytest.raises(InputValueDuplicateSequenceAddress):
#         InputValue(
#             parameter=p1,
#             sequences=[ValueSequence(values=[1, 2, 3]), ValueSequence(values=[4, 5, 6])],
#         )



# import pytest

# from hpcflow.errors import (
#     MissingInputs,
#     TaskTemplateInvalidNesting,
#     TaskTemplateUnexpectedInput,
#     TaskTemplateMultipleInputValues,
#     TaskTemplateMultipleSchemaObjectives,
# )
# from hpcflow.parameters import InputValue, Parameter
# from hpcflow.task import TaskTemplate
# from hpcflow.task_schema import TaskObjective, TaskSchema


# def test_raise_on_multiple_schema_objectives():

#     generate_RVE_schema = TaskSchema(objective=TaskObjective("generate_volume_element"))
#     simulate_schema = TaskSchema(objective=TaskObjective("simulate"))
#     with pytest.raises(TaskTemplateMultipleSchemaObjectives):
#         TaskTemplate(schema=[generate_RVE_schema, simulate_schema])


# def test_raise_on_unexpected_inputs():

#     grid_size = Parameter("grid_size")
#     unknown_input = Parameter("unknown_input")

#     generate_RVE_schema = TaskSchema(
#         objective=TaskObjective("generate_volume_element"), inputs=[grid_size]
#     )

#     with pytest.raises(TaskTemplateUnexpectedInput):
#         TaskTemplate(
#             schema=generate_RVE_schema,
#             input_values=[
#                 InputValue(grid_size, value=[8, 8, 8]),
#                 InputValue(unknown_input, value=4),
#             ],
#         )


# def test_raise_on_multiple_input_values():

#     grid_size = Parameter("grid_size")

#     generate_RVE_schema = TaskSchema(
#         objective=TaskObjective("generate_volume_element"), inputs=[grid_size]
#     )

#     with pytest.raises(TaskTemplateMultipleInputValues):
#         TaskTemplate(
#             schema=generate_RVE_schema,
#             input_values=[
#                 InputValue(grid_size, value=[8, 8, 8]),
#                 InputValue(grid_size, value=7),
#             ],
#         )


# def test_expected_return_defined_and_undefined_input_types():

#     generate_RVE = TaskObjective("generate_volume_element")
#     grid_size = Parameter("grid_size")
#     size = Parameter("size")
#     generate_RVE_schema = TaskSchema(objective=generate_RVE, inputs=[grid_size, size])

#     task = TaskTemplate(
#         schema=generate_RVE_schema, input_values=[InputValue(grid_size, value=[8, 8, 8])]
#     )

#     assert task.defined_input_types == {"grid_size"} and task.undefined_input_types == {
#         "size"
#     }


# def test_expected_return_all_schema_input_types_single_schema():

#     generate_RVE = TaskObjective("generate_volume_element")
#     grid_size = Parameter("grid_size")
#     size = Parameter("size")
#     generate_RVE_schema = TaskSchema(objective=generate_RVE, inputs=[grid_size, size])

#     task = TaskTemplate(schema=generate_RVE_schema)
#     assert task.all_schema_input_types == {"grid_size", "size"}


# def test_expected_return_all_schema_input_types_multiple_schemas():

#     generate_RVE = TaskObjective("generate_volume_element")
#     grid_size = Parameter("grid_size")
#     size = Parameter("size")
#     buffer_layer = Parameter("buffer_layer")

#     generate_RVE_schema_1 = TaskSchema(objective=generate_RVE, inputs=[grid_size, size])
#     generate_RVE_schema_2 = TaskSchema(
#         objective=generate_RVE, inputs=[grid_size, buffer_layer]
#     )

#     task = TaskTemplate(schema=[generate_RVE_schema_1, generate_RVE_schema_2])
#     assert task.all_schema_input_types == {"grid_size", "size", "buffer_layer"}


# def test_expected_unique_name_single_schema():
#     schema = TaskSchema(objective=TaskObjective("simulate"))
#     task = TaskTemplate(schema)
#     assert task.unique_name == "simulate"


# def test_expected_unique_name_single_schema_with_method():
#     schema = TaskSchema(objective=TaskObjective("simulate"), method="method1")
#     task = TaskTemplate(schema)
#     assert task.unique_name == "simulate_method1"


# def test_expected_unique_name_single_schema_with_implementation():
#     schema = TaskSchema(objective=TaskObjective("simulate"), implementation="i1")
#     task = TaskTemplate(schema)
#     assert task.unique_name == "simulate_i1"


# def test_expected_unique_name_single_schema_with_method_and_implementation():
#     schema = TaskSchema(
#         objective=TaskObjective("simulate"), method="method1", implementation="i1"
#     )
#     task = TaskTemplate(schema)
#     assert task.unique_name == "simulate_method1_i1"


# def test_expected_unique_name_multiple_schemas():
#     objective = TaskObjective("simulate")
#     schema_1 = TaskSchema(objective)
#     schema_2 = TaskSchema(objective)
#     task = TaskTemplate([schema_1, schema_2])
#     assert task.unique_name == "simulate"


# def test_expected_unique_name_two_schemas_first_with_method():
#     objective = TaskObjective("simulate")
#     schema_1 = TaskSchema(objective, method="m1")
#     schema_2 = TaskSchema(objective)
#     task = TaskTemplate([schema_1, schema_2])
#     assert task.unique_name == "simulate_m1"


# def test_expected_unique_name_two_schemas_first_with_method_and_implementation():
#     objective = TaskObjective("simulate")
#     schema_1 = TaskSchema(objective, method="m1", implementation="i1")
#     schema_2 = TaskSchema(objective)
#     task = TaskTemplate([schema_1, schema_2])
#     assert task.unique_name == "simulate_m1_i1"


# def test_expected_unique_name_two_schemas_both_with_method():
#     objective = TaskObjective("simulate")
#     schema_1 = TaskSchema(objective, method="m1")
#     schema_2 = TaskSchema(objective, method="m2")
#     task = TaskTemplate([schema_1, schema_2])
#     assert task.unique_name == "simulate_m1_and_m2"


# def test_expected_unique_name_two_schemas_first_with_method_second_with_implementation():
#     objective = TaskObjective("simulate")
#     schema_1 = TaskSchema(objective, method="m1")
#     schema_2 = TaskSchema(objective, implementation="i2")
#     task = TaskTemplate([schema_1, schema_2])
#     assert task.unique_name == "simulate_m1_and_i2"


# def test_expected_unique_name_two_schemas_first_with_implementation_second_with_method():
#     objective = TaskObjective("simulate")
#     schema_1 = TaskSchema(objective, implementation="i1")
#     schema_2 = TaskSchema(objective, method="m2")
#     task = TaskTemplate([schema_1, schema_2])
#     assert task.unique_name == "simulate_i1_and_m2"


# def test_expected_unique_name_two_schemas_both_with_method_and_implementation():
#     objective = TaskObjective("simulate")
#     schema_1 = TaskSchema(objective, method="m1", implementation="i1")
#     schema_2 = TaskSchema(objective, method="m2", implementation="i2")
#     task = TaskTemplate([schema_1, schema_2])
#     assert task.unique_name == "simulate_m1_i1_and_m2_i2"


# def test_resolve_elements():
#     p1 = Parameter("p1")
#     schema = TaskSchema("simulate", inputs=[p1])
#     task = TaskTemplate(schema, input_values=[InputValue(p1, 1)])
#     task.resolve_elements()


# def test_resolve_elements_raise_on_missing_inputs():
#     schema = TaskSchema("simulate", inputs=[Parameter("p1")])
#     task = TaskTemplate(schema)
#     with pytest.raises(MissingInputs):
#         task.resolve_elements()


# def test_raise_on_negative_nesting_order():
#     schema = TaskSchema("simulate", inputs=[Parameter("p1")])
#     with pytest.raises(TaskTemplateInvalidNesting):
#         TaskTemplate(schema, nesting_order={"p1": -1})



# import pytest

# from hpcflow.actions import Action
# from hpcflow.command_files import FileSpec, InputFileGenerator
# from hpcflow.commands import Command
# from hpcflow.errors import MissingActionEnvironment
# from hpcflow.parameters import Parameter


# @pytest.fixture
# def dummy_commands_kwargs():
#     return {"commands": [Command("ls")]}


# @pytest.fixture
# def dummy_commands_spec():
#     return {"commands": [{"command": "ls"}]}


# @pytest.fixture
# def dummy_action_kwargs_no_env(dummy_commands_kwargs):
#     act_kwargs = {**dummy_commands_kwargs}
#     return act_kwargs


# @pytest.fixture
# def dummy_action_spec_no_env(dummy_commands_spec):
#     act_spec = {**dummy_commands_spec}
#     return act_spec


# @pytest.fixture
# def dummy_action_kwargs_pre_proc():
#     act_kwargs = {
#         "commands": [Command("ls")],
#         "input_file_generators": [
#             InputFileGenerator(
#                 input_file=FileSpec("inp_file", name="file.inp"), inputs=[Parameter("p1")]
#             )
#         ],
#     }
#     return act_kwargs


# def test_raise_on_no_envs(dummy_action_kwargs_no_env):
#     with pytest.raises(TypeError):
#         Action(**dummy_action_kwargs_no_env)


# def test_spec_raise_on_no_envs(dummy_action_spec_no_env):
#     with pytest.raises(MissingActionEnvironment):
#         Action.from_spec(dummy_action_spec_no_env)


# def test_1(dummy_action_kwargs_pre_proc):
#     act = Action(**dummy_action_kwargs_pre_proc)



from hpcflow import __version__
from hpcflow.sdk import ConfigOptions
from hpcflow.sdk.app import BaseApp

config_options = ConfigOptions(
    directory_env_var="HPCFLOW_CONFIG_DIR",
    default_directory="~/.hpcflow",
    sentry_DSN="https://2463b288fd1a40f4bada9f5ff53f6811@o1180430.ingest.sentry.io/6293231",
    sentry_traces_sample_rate=1.0,
    sentry_env="main" if "a" in __version__ else "develop",
)

hpcflow = BaseApp(
    name="hpcflow",
    version=__version__,
    description="Computational workflow management",
    config_options=config_options,
    pytest_args=[
        "--verbose",
        "--exitfirst",
    ],
)

load_config = hpcflow.load_config
reload_config = hpcflow.reload_config
make_workflow = hpcflow.make_workflow

# expose core classes that require access to the App instance:
TaskSchema = hpcflow.TaskSchema
Task = hpcflow.Task
WorkflowTask = hpcflow.WorkflowTask
Workflow = hpcflow.Workflow
WorkflowTemplate = hpcflow.WorkflowTemplate
Action = hpcflow.Action
ActionScope = hpcflow.ActionScope
ActionScopeType = hpcflow.ActionScopeType
Environment = hpcflow.Environment
InputFile = hpcflow.InputFile
InputSource = hpcflow.InputSource
InputSourceType = hpcflow.InputSourceType
InputSourceMode = hpcflow.InputSourceMode
Command = hpcflow.Command
ActionEnvironment = hpcflow.ActionEnvironment
Parameter = hpcflow.Parameter
ValueSequence = hpcflow.ValueSequence
ZarrEncodable = hpcflow.ZarrEncodable
InputValue = hpcflow.InputValue
ResourceSpec = hpcflow.ResourceSpec
SchemaInput = hpcflow.SchemaInput
SchemaOutput = hpcflow.SchemaOutput
TaskSourceType = hpcflow.TaskSourceType



"""API functions, which are dynamically added to the BaseApp class on __init__"""

import importlib
from hpcflow.sdk.core.utils import load_config


@load_config
def make_workflow(app, dir):
    """make a new {name} workflow.

    Parameters
    ----------
    dir
        Directory into which the workflow will be generated.

    Returns
    -------
    nonsense : Workflow

    """
    pass
    app.API_logger.info("hey")


def run_hpcflow_tests(app, *args):
    """Run hpcflow test suite. This function is only available from derived apps.

    Notes
    -----
    It may not be possible to run hpcflow tests after/before running tests of the derived
    app within the same process, due to caching."""

    from hpcflow.api import hpcflow

    hpcflow.run_tests(*args)


def run_tests(app, *args):
    """Run {name} test suite."""

    try:
        import pytest
    except ModuleNotFoundError:
        raise RuntimeError(f"{app.name} has not been built with testing dependencies.")

    with importlib.resources.path(app.name, "tests") as test_dir:
        pytest.main(
            [str(test_dir)]
            + (app.pytest_args or [])
            + list(args)
            + ["--log-level", "INFO"]
        )



# import pytest

# from click.testing import CliRunner

# from hpcflow import __version__
# from hpcflow.cli.cli import cli


# def test_version():
#     runner = CliRunner()
#     result = runner.invoke(cli, args="--version")
#     assert result.output.strip() == f"hpcflow, version {__version__}"
